@@ -7,33 +7,26 @@
 
 #include "Rvector_utils.h"
 #include "coerceVector2.h"  /* for _CoercionWarning() */
-#include "leaf_vector_utils.h"
+#include "leaf_utils.h"
 
 #include <limits.h>  /* for INT_MAX */
-#include <string.h>  /* for memset() */
+#include <string.h>  /* for strcmp() and memcpy() */
 
 
-/* General purpose copy function. */
-static inline int copy_Rvector_elts(
-		SEXP in,  R_xlen_t in_offset,
-		SEXP out, R_xlen_t out_offset,
-		R_xlen_t nelt)
+static void copy_nzvals_elts_to_Rsubvec(SEXP nzvals,
+		SEXP out, int out_offset, int nelt)
 {
-	SEXPTYPE Rtype;
-	CopyRVectorElts_FUNType copy_Rvector_elts_FUN;
-
-	Rtype = TYPEOF(in);
-	copy_Rvector_elts_FUN = _select_copy_Rvector_elts_FUN(Rtype);
-	if (copy_Rvector_elts_FUN == NULL)
-		return -1;
-	if (TYPEOF(out) != Rtype)
-		return -1;
-	if (in_offset  + nelt > XLENGTH(in))
-		return -1;
-	if (out_offset + nelt > XLENGTH(out))
-		return -1;
-	copy_Rvector_elts_FUN(in, in_offset, out, out_offset, nelt);
-	return 0;
+	if (nzvals == R_NilValue) {
+		/* lacunar leaf */
+		_set_Rsubvec_elts_to_one(out, (R_xlen_t) out_offset,
+					 (R_xlen_t) nelt);
+	} else {
+		/* regular leaf */
+		_copy_Rvector_elts(nzvals, 0,
+				   out, (R_xlen_t) out_offset,
+				   (R_xlen_t) nelt);
+	}
+	return;
 }
 
 
@@ -41,40 +34,46 @@ static inline int copy_Rvector_elts(
  * type() setter
  */
 
-/* Recursive. */
-static int REC_set_SVT_type(SEXP SVT, const int *dim, int ndim,
-		SEXPTYPE new_Rtype, int *warn, int *offs_buf)
+/* Returns 1 if coercion to the requested type turned 'leaf' into an
+   R_NilValue, and 0 otherwise. */
+static int INPLACE_modify_leaf_type(SEXP leaf,
+		SEXPTYPE new_Rtype, int *warn, int *nzoffs_buf)
 {
-	SEXP new_lv, subSVT;
-	int SVT_len, is_empty, i, ret;
+	SEXP new_leaf = _coerce_leaf(leaf, new_Rtype, warn, nzoffs_buf);
+	if (new_leaf == R_NilValue)
+		return 1;
+	PROTECT(new_leaf);
+	replace_leaf_nzvals(leaf, get_leaf_nzvals(new_leaf));
+	replace_leaf_nzoffs(leaf, get_leaf_nzoffs(new_leaf));
+	UNPROTECT(1);
+	return 0;
+}
 
+/* Recursive. */
+static int REC_INPLACE_modify_SVT_type(SEXP SVT, const int *dim, int ndim,
+		SEXPTYPE new_Rtype, int *warn, int *nzoffs_buf)
+{
 	if (SVT == R_NilValue)
 		return 1;
 
 	if (ndim == 1) {
-		/* 'SVT' is a "leaf vector". */
-		new_lv = _coerce_leaf_vector(SVT, new_Rtype, warn, offs_buf);
-		if (new_lv == R_NilValue)
-			return 1;
-		PROTECT(new_lv);
-		SET_VECTOR_ELT(SVT, 0, VECTOR_ELT(new_lv, 0));
-		SET_VECTOR_ELT(SVT, 1, VECTOR_ELT(new_lv, 1));
-		UNPROTECT(1);
-		return 0;
+		/* 'SVT' is a leaf (i.e. a 1D SVT). */
+		return INPLACE_modify_leaf_type(SVT, new_Rtype,
+						warn, nzoffs_buf);
 	}
 
 	/* 'SVT' is a regular node (list). */
-	SVT_len = LENGTH(SVT);
+	int SVT_len = LENGTH(SVT);
 
 	/* Sanity check (should never fail). */
 	if (SVT_len != dim[ndim - 1])
 		return -1;
 
-	is_empty = 1;
-	for (i = 0; i < SVT_len; i++) {
-		subSVT = VECTOR_ELT(SVT, i);
-		ret = REC_set_SVT_type(subSVT, dim, ndim - 1,
-				       new_Rtype, warn, offs_buf);
+	int is_empty = 1;
+	for (int i = 0; i < SVT_len; i++) {
+		SEXP subSVT = VECTOR_ELT(SVT, i);
+		int ret = REC_INPLACE_modify_SVT_type(subSVT, dim, ndim - 1,
+						   new_Rtype, warn, nzoffs_buf);
 		if (ret < 0)
 			return -1;
 		if (ret == 1) {
@@ -90,32 +89,29 @@ static int REC_set_SVT_type(SEXP SVT, const int *dim, int ndim,
 SEXP C_set_SVT_SparseArray_type(SEXP x_dim, SEXP x_type, SEXP x_SVT,
 		SEXP new_type)
 {
-	SEXPTYPE x_Rtype, new_Rtype;
-	int warn, *offs_buf, ret;
-	SEXP ans;
-
-	x_Rtype = _get_Rtype_from_Rstring(x_type);
+	SEXPTYPE x_Rtype = _get_Rtype_from_Rstring(x_type);
 	if (x_Rtype == 0)
 		error("SparseArray internal error in "
 		      "C_set_SVT_SparseArray_type():\n"
 		      "    invalid 'x_type' value");
-	new_Rtype = _get_Rtype_from_Rstring(new_type);
+	SEXPTYPE new_Rtype = _get_Rtype_from_Rstring(new_type);
 	if (new_Rtype == 0)
 		error("invalid supplied type");
 
 	if (new_Rtype == x_Rtype || x_SVT == R_NilValue)
 		return x_SVT;
 
-	offs_buf = (int *) R_alloc(INTEGER(x_dim)[0], sizeof(int));
-	ans = PROTECT(duplicate(x_SVT));
-	warn = 0;
-	ret = REC_set_SVT_type(ans, INTEGER(x_dim), LENGTH(x_dim),
-			       new_Rtype, &warn, offs_buf);
+	int *nzoffs_buf = (int *) R_alloc(INTEGER(x_dim)[0], sizeof(int));
+	SEXP ans = PROTECT(duplicate(x_SVT));
+	int warn = 0;
+	int ret = REC_INPLACE_modify_SVT_type(ans,
+				INTEGER(x_dim), LENGTH(x_dim),
+				new_Rtype, &warn, nzoffs_buf);
 	if (ret < 0) {
 		UNPROTECT(1);
 		error("SparseArray internal error in "
 		      "C_set_SVT_SparseArray_type():\n"
-		      "    REC_set_SVT_type() returned an error");
+		      "    REC_INPLACE_modify_SVT_type() returned an error");
 	}
 	if (warn)
 		_CoercionWarning(warn);
@@ -129,18 +125,16 @@ SEXP C_set_SVT_SparseArray_type(SEXP x_dim, SEXP x_type, SEXP x_SVT,
 SEXP _coerce_SVT(SEXP SVT, const int *dim, int ndim,
 		 SEXPTYPE from_Rtype, SEXPTYPE to_Rtype, int *offs_buf)
 {
-	SEXP ans;
-	int ret, warn;
-
         if (from_Rtype == to_Rtype)
 		return SVT;
-	ans = PROTECT(duplicate(SVT));
-	ret = REC_set_SVT_type(ans, dim, ndim,
-			       to_Rtype, &warn, offs_buf);
+	SEXP ans = PROTECT(duplicate(SVT));
+	int warn;
+	int ret = REC_INPLACE_modify_SVT_type(ans, dim, ndim,
+					      to_Rtype, &warn, offs_buf);
 	if (ret < 0) {
 		UNPROTECT(1);
 		error("SparseArray internal error in _coerce_SVT():\n"
-		      "    REC_set_SVT_type() returned an error");
+		      "    REC_INPLACE_modify_SVT_type() returned an error");
 	}
 	UNPROTECT(1);
 	return ret == 1 ? R_NilValue : ans;
@@ -154,23 +148,19 @@ SEXP _coerce_SVT(SEXP SVT, const int *dim, int ndim,
 /* Recursive. */
 R_xlen_t _REC_nzcount_SVT(SEXP SVT, int ndim)
 {
-	R_xlen_t nzcount;
-	int SVT_len, i;
-	SEXP subSVT;
-
 	if (SVT == R_NilValue)
 		return 0;
 
 	if (ndim == 1) {
-		/* 'SVT' is a "leaf vector". */
-		return XLENGTH(VECTOR_ELT(SVT, 0));
+		/* 'SVT' is a leaf (i.e. a 1D SVT). */
+		return (R_xlen_t) get_leaf_nzcount(SVT);
 	}
 
 	/* 'SVT' is a regular node (list). */
-	nzcount = 0;
-	SVT_len = LENGTH(SVT);
-	for (i = 0; i < SVT_len; i++) {
-		subSVT = VECTOR_ELT(SVT, i);
+	R_xlen_t nzcount = 0;
+	int SVT_len = LENGTH(SVT);
+	for (int i = 0; i < SVT_len; i++) {
+		SEXP subSVT = VECTOR_ELT(SVT, i);
 		nzcount += _REC_nzcount_SVT(subSVT, ndim - 1);
 	}
 	return nzcount;
@@ -179,9 +169,7 @@ R_xlen_t _REC_nzcount_SVT(SEXP SVT, int ndim)
 /* --- .Call ENTRY POINT --- */
 SEXP C_nzcount_SVT_SparseArray(SEXP x_dim, SEXP x_SVT)
 {
-	R_xlen_t nzcount;
-
-	nzcount = _REC_nzcount_SVT(x_SVT, LENGTH(x_dim));
+	R_xlen_t nzcount = _REC_nzcount_SVT(x_SVT, LENGTH(x_dim));
 	if (nzcount > INT_MAX)
 		return ScalarReal((double) nzcount);
 	return ScalarInteger((int) nzcount);
@@ -195,9 +183,7 @@ SEXP C_nzcount_SVT_SparseArray(SEXP x_dim, SEXP x_SVT)
 static inline void from_offs_to_int_Lindex(const int *offs, int n,
 		int arr_offset, int *Lindex)
 {
-	int k;
-
-	for (k = 0; k < n; k++)
+	for (int k = 0; k < n; k++)
 		Lindex[k] = arr_offset + offs[k] + 1;  /* will never overflow */
 	return;
 }
@@ -205,9 +191,7 @@ static inline void from_offs_to_int_Lindex(const int *offs, int n,
 static inline void from_offs_to_double_Lindex(const int *offs, int n,
 		R_xlen_t arr_offset, double *Lindex)
 {
-	int k;
-
-	for (k = 0; k < n; k++)
+	for (int k = 0; k < n; k++)
 		Lindex[k] = (double) (arr_offset + offs[k] + 1);
 	return;
 }
@@ -217,42 +201,39 @@ static int REC_nzwhich_SVT_as_Lindex(SEXP SVT,
 		const int *dim, const R_xlen_t *dimcumprod, int ndim,
 		R_xlen_t arr_offset, SEXP Lindex, R_xlen_t *Lindex_offset)
 {
-	int lv_len, SVT_len, i, ret;
-	SEXP lv_offs, lv_vals, subSVT;
-	R_xlen_t subarr_len;
-
 	if (SVT == R_NilValue)
 		return 0;
 
 	if (ndim == 1) {
-		/* 'SVT' is a "leaf vector". */
-		lv_len = _split_leaf_vector(SVT, &lv_offs, &lv_vals);
-		if (lv_len < 0)
+		/* 'SVT' is a leaf (i.e. 1D SVT). */
+		SEXP nzvals, nzoffs;
+		int nzcount = unzip_leaf(SVT, &nzvals, &nzoffs);
+		if (nzcount < 0)
 			return -1;
 		if (IS_INTEGER(Lindex)) {
-			from_offs_to_int_Lindex(INTEGER(lv_offs), lv_len,
+			from_offs_to_int_Lindex(INTEGER(nzoffs), nzcount,
 					(int) arr_offset,
 					INTEGER(Lindex) + *Lindex_offset);
 		} else {
-			from_offs_to_double_Lindex(INTEGER(lv_offs), lv_len,
+			from_offs_to_double_Lindex(INTEGER(nzoffs), nzcount,
 					arr_offset,
 					REAL(Lindex) + *Lindex_offset);
 		}
-		(*Lindex_offset) += lv_len;
+		(*Lindex_offset) += nzcount;
 		return 0;
 	}
 
 	/* 'SVT' is a regular node (list). */
-	SVT_len = LENGTH(SVT);
+	int SVT_len = LENGTH(SVT);
 
 	/* Sanity check (should never fail). */
 	if (SVT_len != dim[ndim - 1])
 		return -1;
 
-	subarr_len = dimcumprod[ndim - 2];
-	for (i = 0; i < SVT_len; i++) {
-		subSVT = VECTOR_ELT(SVT, i);
-		ret = REC_nzwhich_SVT_as_Lindex(subSVT,
+	R_xlen_t subarr_len = dimcumprod[ndim - 2];
+	for (int i = 0; i < SVT_len; i++) {
+		SEXP subSVT = VECTOR_ELT(SVT, i);
+		int ret = REC_nzwhich_SVT_as_Lindex(subSVT,
 					dim, dimcumprod, ndim - 1,
 					arr_offset, Lindex, Lindex_offset);
 		if (ret < 0)
@@ -265,25 +246,18 @@ static int REC_nzwhich_SVT_as_Lindex(SEXP SVT,
 static SEXP nzwhich_SVT_as_Lindex(SEXP SVT, const int *dim, int ndim,
 		R_xlen_t nzcount)
 {
-	R_xlen_t *dimcumprod, p, Lindex_offset;
-	int along, ret;
-	SEXP ans;
-
-	dimcumprod = (R_xlen_t *) R_alloc(ndim, sizeof(R_xlen_t));
-	p = 1;
-	for (along = 0; along < ndim; along++) {
+	R_xlen_t *dimcumprod = (R_xlen_t *) R_alloc(ndim, sizeof(R_xlen_t));
+	R_xlen_t p = 1;
+	for (int along = 0; along < ndim; along++) {
 		p *= dim[along];
 		dimcumprod[along] = p;
 	}
 
-	if (p > INT_MAX) {
-		ans = PROTECT(NEW_NUMERIC(nzcount));
-	} else {
-		ans = PROTECT(NEW_INTEGER(nzcount));
-	}
-	Lindex_offset = 0;
-	ret = REC_nzwhich_SVT_as_Lindex(SVT, dim, dimcumprod, ndim,
-					0, ans, &Lindex_offset);
+	SEXPTYPE ans_Rtype = p > INT_MAX ? REALSXP : INTSXP;
+	SEXP ans = PROTECT(allocVector(ans_Rtype, nzcount));
+	R_xlen_t Lindex_offset = 0;
+	int ret = REC_nzwhich_SVT_as_Lindex(SVT, dim, dimcumprod, ndim,
+					    0, ans, &Lindex_offset);
 	UNPROTECT(1);
 	if (ret < 0)
 		error("SparseArray internal error in "
@@ -301,52 +275,45 @@ static SEXP nzwhich_SVT_as_Lindex(SEXP SVT, const int *dim, int ndim,
 
 /* Recursive. */
 static int REC_extract_nzcoo_and_nzvals_from_SVT(SEXP SVT,
-		int *nzcoo, int nzcoo_nrow, int nzcoo_ncol,
+		int *out_nzcoo, int nzcoo_nrow, int nzcoo_ncol,
 		int *rowbuf, int rowbuf_offset,
-		SEXP nzvals, int *nzvals_offset)
+		SEXP out_nzvals, int *nzvals_offset)
 {
-	int SVT_len, i, ret, lv_len, k, *p, j;
-	SEXP subSVT, lv_offs, lv_vals;
-
 	if (SVT == R_NilValue)
 		return 0;
 
 	if (rowbuf_offset > 0) {
 		if (!isVectorList(SVT))  // IS_LIST() is broken
 			return -1;
-		SVT_len = LENGTH(SVT);
-		for (i = 0; i < SVT_len; i++) {
-			subSVT = VECTOR_ELT(SVT, i);
+		int SVT_len = LENGTH(SVT);
+		for (int i = 0; i < SVT_len; i++) {
+			SEXP subSVT = VECTOR_ELT(SVT, i);
 			rowbuf[rowbuf_offset] = i + 1;
-			ret = REC_extract_nzcoo_and_nzvals_from_SVT(subSVT,
-					nzcoo, nzcoo_nrow, nzcoo_ncol,
+			int ret = REC_extract_nzcoo_and_nzvals_from_SVT(subSVT,
+					out_nzcoo, nzcoo_nrow, nzcoo_ncol,
 					rowbuf, rowbuf_offset - 1,
-					nzvals, nzvals_offset);
+					out_nzvals, nzvals_offset);
 			if (ret < 0)
 				return -1;
 		}
 		return 0;
 	}
 
-	/* 'SVT' is a "leaf vector". */
-	lv_len = _split_leaf_vector(SVT, &lv_offs, &lv_vals);
-	if (lv_len < 0)
-		return -1;
+	/* 'SVT' is a leaf (i.e. a 1D SVT). */
+	SEXP leaf_nzvals, leaf_nzoffs;
+	int leaf_nzcount = unzip_leaf(SVT, &leaf_nzvals, &leaf_nzoffs);
 
-	if (nzvals != R_NilValue) {
-		ret = copy_Rvector_elts(lv_vals, (R_xlen_t) 0,
-					nzvals, (R_xlen_t) *nzvals_offset,
-					XLENGTH(lv_vals));
-		if (ret < 0)
-			return -1;
-	}
+	if (out_nzvals != R_NilValue)
+		copy_nzvals_elts_to_Rsubvec(leaf_nzvals,
+					    out_nzvals, *nzvals_offset,
+					    leaf_nzcount);
 
-	for (k = 0; k < lv_len; k++) {
-		rowbuf[0] = INTEGER(lv_offs)[k] + 1;
+	for (int k = 0; k < leaf_nzcount; k++) {
+		rowbuf[0] = INTEGER(leaf_nzoffs)[k] + 1;
 
 		/* Copy 'rowbuf' to 'nzcoo'. */
-		p = nzcoo + *nzvals_offset;
-		for (j = 0; j < nzcoo_ncol; j++) {
+		int *p = out_nzcoo + *nzvals_offset;
+		for (int j = 0; j < nzcoo_ncol; j++) {
 			*p = rowbuf[j];
 			p += nzcoo_nrow;
 		}
@@ -358,19 +325,16 @@ static int REC_extract_nzcoo_and_nzvals_from_SVT(SEXP SVT,
 
 static SEXP extract_nzcoo_and_nzvals_from_SVT(SEXP SVT,
 		int nzcoo_nrow, int nzcoo_ncol,
-		SEXP nzvals)
+		SEXP out_nzvals)
 {
-	int *rowbuf, nzvals_offset, ret;
-	SEXP nzcoo;
+	int *rowbuf = (int *) R_alloc(nzcoo_ncol, sizeof(int));
+	SEXP out_nzcoo = PROTECT(allocMatrix(INTSXP, nzcoo_nrow, nzcoo_ncol));
 
-	rowbuf = (int *) R_alloc(nzcoo_ncol, sizeof(int));
-	nzcoo = PROTECT(allocMatrix(INTSXP, nzcoo_nrow, nzcoo_ncol));
-
-	nzvals_offset = 0;
-	ret = REC_extract_nzcoo_and_nzvals_from_SVT(SVT,
-			INTEGER(nzcoo), nzcoo_nrow, nzcoo_ncol,
+	int nzvals_offset = 0;
+	int ret = REC_extract_nzcoo_and_nzvals_from_SVT(SVT,
+			INTEGER(out_nzcoo), nzcoo_nrow, nzcoo_ncol,
 			rowbuf, nzcoo_ncol - 1,
-			nzvals, &nzvals_offset);
+			out_nzvals, &nzvals_offset);
 	if (ret < 0) {
 		UNPROTECT(1);
 		error("SparseArray internal error in "
@@ -387,17 +351,14 @@ static SEXP extract_nzcoo_and_nzvals_from_SVT(SEXP SVT,
 	}
 
 	UNPROTECT(1);
-	return nzcoo;
+	return out_nzcoo;
 }
 
 /* --- .Call ENTRY POINT --- */
 SEXP C_nzwhich_SVT_SparseArray(SEXP x_dim, SEXP x_SVT, SEXP arr_ind)
 {
-	int x_ndim;
-	R_xlen_t nzcount;
-
-	x_ndim = LENGTH(x_dim);
-	nzcount = _REC_nzcount_SVT(x_SVT, x_ndim);
+	int x_ndim = LENGTH(x_dim);
+	R_xlen_t nzcount = _REC_nzcount_SVT(x_SVT, x_ndim);
 
 	if (!LOGICAL(arr_ind)[0]) {
 		/* Return linear indices of nonzero array elements in an
@@ -422,32 +383,30 @@ SEXP C_nzwhich_SVT_SparseArray(SEXP x_dim, SEXP x_SVT, SEXP arr_ind)
  */
 
 /* Recursive. */
-static int REC_dump_SVT_to_Rsubarray(SEXP SVT,
+static int REC_unroll_SVT_into_Rarray(SEXP SVT,
 		const int *dim, int ndim,
 		SEXP Rarray, R_xlen_t arr_offset, R_xlen_t subarr_len)
 {
-	int SVT_len, i, ret;
-	SEXP subSVT;
-
 	if (SVT == R_NilValue)
 		return 0;
 
 	if (ndim == 1) {
-		/* 'SVT' is a "leaf vector". */
-		return _expand_leaf_vector(SVT, Rarray, arr_offset);
+		/* 'SVT' is a leaf (i.e. a 1D SVT). */
+		_expand_leaf(SVT, Rarray, arr_offset);
+		return 0;
 	}
 
 	/* 'SVT' is a regular node (list). */
-	SVT_len = LENGTH(SVT);
+	int SVT_len = LENGTH(SVT);
 
 	/* Sanity check (should never fail). */
 	if (SVT_len != dim[ndim - 1])
 		return -1;
 
 	subarr_len /= SVT_len;
-	for (i = 0; i < SVT_len; i++) {
-		subSVT = VECTOR_ELT(SVT, i);
-		ret = REC_dump_SVT_to_Rsubarray(subSVT,
+	for (int i = 0; i < SVT_len; i++) {
+		SEXP subSVT = VECTOR_ELT(SVT, i);
+		int ret = REC_unroll_SVT_into_Rarray(subSVT,
 				dim, ndim - 1,
 				Rarray, arr_offset, subarr_len);
 		if (ret < 0)
@@ -461,18 +420,14 @@ static int REC_dump_SVT_to_Rsubarray(SEXP SVT,
 SEXP C_from_SVT_SparseArray_to_Rarray(SEXP x_dim, SEXP x_dimnames,
 		SEXP x_type, SEXP x_SVT)
 {
-	SEXPTYPE Rtype;
-	SEXP ans;
-	int ret;
-
-	Rtype = _get_Rtype_from_Rstring(x_type);
+	SEXPTYPE Rtype = _get_Rtype_from_Rstring(x_type);
 	if (Rtype == 0)
 		error("SparseArray internal error in "
 		      "C_from_SVT_SparseArray_to_Rarray():\n"
 		      "    SVT_SparseArray object has invalid type");
 
-	ans = PROTECT(_new_Rarray0(Rtype, x_dim, x_dimnames));
-	ret = REC_dump_SVT_to_Rsubarray(x_SVT,
+	SEXP ans = PROTECT(_new_Rarray0(Rtype, x_dim, x_dimnames));
+	int ret = REC_unroll_SVT_into_Rarray(x_SVT,
 				INTEGER(x_dim), LENGTH(x_dim),
 				ans, 0, XLENGTH(ans));
 	UNPROTECT(1);
@@ -494,32 +449,28 @@ static SEXP REC_build_SVT_from_Rsubarray(
 		const int *dim, int ndim,
 		SEXPTYPE ans_Rtype, int *warn, int *offs_buf)
 {
-	SEXP ans, ans_elt;
-	int SVT_len, is_empty, i;
-
 	if (ndim == 1) {
 		/* Sanity check (should never fail). */
 		if (dim[0] != subarr_len)
 			error("SparseArray internal error in "
 			      "REC_build_SVT_from_Rsubarray():\n"
 			      "    dim[0] != subarr_len");
-		ans = _make_leaf_vector_from_Rsubvec(
-					Rarray, arr_offset, dim[0],
-					offs_buf, 1);
+		SEXP ans = _make_leaf_from_Rsubvec(Rarray, arr_offset, dim[0],
+						   offs_buf, 1);
 		if (ans_Rtype == TYPEOF(Rarray) || ans == R_NilValue)
 			return ans;
 		PROTECT(ans);
-		ans = _coerce_leaf_vector(ans, ans_Rtype, warn, offs_buf);
+		ans = _coerce_leaf(ans, ans_Rtype, warn, offs_buf);
 		UNPROTECT(1);
 		return ans;
 	}
 
-	SVT_len = dim[ndim - 1];  /* cannot be 0 so safe to divide below */
+	int SVT_len = dim[ndim - 1];  /* cannot be 0 so safe to divide below */
 	subarr_len /= SVT_len;
-	ans = PROTECT(NEW_LIST(SVT_len));
-	is_empty = 1;
-	for (i = 0; i < SVT_len; i++) {
-		ans_elt = REC_build_SVT_from_Rsubarray(
+	SEXP ans = PROTECT(NEW_LIST(SVT_len));
+	int is_empty = 1;
+	for (int i = 0; i < SVT_len; i++) {
+		SEXP ans_elt = REC_build_SVT_from_Rsubarray(
 					Rarray, arr_offset, subarr_len,
 					dim, ndim - 1,
 					ans_Rtype, warn, offs_buf);
@@ -538,24 +489,19 @@ static SEXP REC_build_SVT_from_Rsubarray(
 /* --- .Call ENTRY POINT --- */
 SEXP C_build_SVT_from_Rarray(SEXP x, SEXP ans_type)
 {
-	SEXPTYPE ans_Rtype;
-	int x_ndim, warn, *offs_buf;
-	R_xlen_t x_len;
-	SEXP x_dim, ans;
-
-	ans_Rtype = _get_Rtype_from_Rstring(ans_type);
+	SEXPTYPE ans_Rtype = _get_Rtype_from_Rstring(ans_type);
 	if (ans_Rtype == 0)
 		error("invalid requested type");
 
-	x_len = XLENGTH(x);
+	R_xlen_t x_len = XLENGTH(x);
 	if (x_len == 0)  /* means that 'any(dim(x) == 0)' is TRUE */
 		return R_NilValue;
 
-	x_dim = GET_DIM(x);  /* does not contain zeros */
-	x_ndim = LENGTH(x_dim);
-	offs_buf = (int *) R_alloc(INTEGER(x_dim)[0], sizeof(int));
-	warn = 0;
-	ans = REC_build_SVT_from_Rsubarray(x, 0, x_len,
+	SEXP x_dim = GET_DIM(x);  /* does not contain zeros */
+	int x_ndim = LENGTH(x_dim);
+	int *offs_buf = (int *) R_alloc(INTEGER(x_dim)[0], sizeof(int));
+	int warn = 0;
+	SEXP ans = REC_build_SVT_from_Rsubarray(x, 0, x_len,
 					   INTEGER(x_dim), x_ndim,
 					   ans_Rtype, &warn, offs_buf);
 	if (warn) {
@@ -570,92 +516,80 @@ SEXP C_build_SVT_from_Rarray(SEXP x, SEXP ans_type)
 
 
 /****************************************************************************
- * Going from SVT_SparseMatrix to [d|l]gCMatrix
+ * Going from SVT_SparseMatrix to [d|l|n]gCMatrix
  */
 
 /* Returns nb of nonzero elements in column. */
-static int dump_col_to_CsparseMatrix_slots(SEXP SVT, int col_idx,
-		SEXP ans_i, SEXP ans_x, int offset)
+static int dump_leaf_to_ix(SEXP leaf,
+		SEXP sloti, SEXP slotx, int ix_offset)
 {
-	SEXP subSVT, lv_offs, lv_vals;
-	int lv_len, ret;
-
-	subSVT = VECTOR_ELT(SVT, col_idx);
-	if (subSVT == R_NilValue)
+	if (leaf == R_NilValue)
 		return 0;
 
-	/* 'subSVT' is a "leaf vector". */
-	lv_len = _split_leaf_vector(subSVT, &lv_offs, &lv_vals);
-	if (lv_len < 0)
-		return -1;
+	SEXP nzvals, nzoffs;
+	int nzcount = unzip_leaf(leaf, &nzvals, &nzoffs);
 
-	/* Copy 0-based row indices from 'lv_offs' to 'ans_i'. */
-	_copy_INTEGER_elts(lv_offs, (R_xlen_t) 0,
-			ans_i, (R_xlen_t) offset,
-			XLENGTH(lv_offs));
+	/* Copy 0-based row indices from 'nzoffs' to 'sloti'. */
+	memcpy(INTEGER(sloti) + ix_offset, INTEGER(nzoffs),
+	       sizeof(int) * nzcount);
 
-	ret = copy_Rvector_elts(lv_vals, (R_xlen_t) 0,
-			ans_x, (R_xlen_t) offset,
-			XLENGTH(lv_vals));
-	if (ret < 0)
-		return -1;
+	if (slotx == R_NilValue)
+		return nzcount;
 
-	return lv_len;
+	/* Copy 'nzvals' to 'slotx'. */
+	copy_nzvals_elts_to_Rsubvec(nzvals, slotx, ix_offset, nzcount);
+	return nzcount;
 }
 
 static int dump_SVT_to_CsparseMatrix_slots(SEXP x_SVT, int x_ncol,
-		SEXP ans_p, SEXP ans_i, SEXP ans_x)
+		SEXP slotp, SEXP sloti, SEXP slotx)
 {
-	int offset, j, nzcount;
-
-	INTEGER(ans_p)[0] = 0;
-	offset = 0;
-	for (j = 0; j < x_ncol; j++) {
-		nzcount = dump_col_to_CsparseMatrix_slots(x_SVT, j,
-						ans_i, ans_x, offset);
+	INTEGER(slotp)[0] = 0;
+	int ix_offset = 0;
+	for (int j = 0; j < x_ncol; j++) {
+		SEXP leaf = VECTOR_ELT(x_SVT, j);
+		int nzcount = dump_leaf_to_ix(leaf, sloti, slotx, ix_offset);
 		if (nzcount < 0)
 			return -1;
-		offset += nzcount;
-		INTEGER(ans_p)[j + 1] = offset;
+		ix_offset += nzcount;
+		INTEGER(slotp)[j + 1] = ix_offset;
 	}
 	return 0;
 }
 
 /* --- .Call ENTRY POINT --- */
 SEXP C_from_SVT_SparseMatrix_to_CsparseMatrix(SEXP x_dim,
-		SEXP x_type, SEXP x_SVT)
+		SEXP x_type, SEXP x_SVT, SEXP as_ngCMatrix)
 {
-	R_xlen_t nzcount;
-	SEXPTYPE x_Rtype;
-	int x_ncol, ret;
-	SEXP ans_p, ans_i, ans_x, ans;
-
 	if (LENGTH(x_dim) != 2)
 		error("object to coerce to [d|l]gCMatrix "
 		      "must have exactly 2 dimensions");
 
-	nzcount = _REC_nzcount_SVT(x_SVT, 2);
-	if (nzcount > INT_MAX)
+	R_xlen_t x_nzcount = _REC_nzcount_SVT(x_SVT, LENGTH(x_dim));
+	if (x_nzcount > INT_MAX)
 		error("SVT_SparseMatrix object contains too many nonzero "
 		      "values to be turned into a dgCMatrix or lgCMatrix "
 		      "object");
 
-	x_Rtype = _get_Rtype_from_Rstring(x_type);
+	SEXPTYPE x_Rtype = _get_Rtype_from_Rstring(x_type);
 	if (x_Rtype == 0)
 		error("SparseArray internal error in "
 		      "C_from_SVT_SparseMatrix_to_CsparseMatrix():\n"
 		      "    SVT_SparseMatrix object has invalid type");
 
-	x_ncol = INTEGER(x_dim)[1];
-
-	ans_i = PROTECT(NEW_INTEGER(nzcount));
-	ans_x = PROTECT(allocVector(x_Rtype, nzcount));
-	if (nzcount == 0) {
-		ans_p = PROTECT(_new_Rvector0(INTSXP, (R_xlen_t) x_ncol + 1));
+	int x_ncol = INTEGER(x_dim)[1];
+	SEXP sloti = PROTECT(NEW_INTEGER(x_nzcount));
+	int drop_slotx = LOGICAL(as_ngCMatrix)[0];
+	SEXP slotx = R_NilValue;
+	if (!drop_slotx)
+		slotx = PROTECT(allocVector(x_Rtype, x_nzcount));
+	SEXP slotp;
+	if (x_SVT == R_NilValue) {
+		slotp = PROTECT(_new_Rvector0(INTSXP, (R_xlen_t) x_ncol + 1));
 	} else {
-		ans_p = PROTECT(NEW_INTEGER(x_ncol + 1));
-		ret = dump_SVT_to_CsparseMatrix_slots(x_SVT, x_ncol,
-						      ans_p, ans_i, ans_x);
+		slotp = PROTECT(NEW_INTEGER(x_ncol + 1));
+		int ret = dump_SVT_to_CsparseMatrix_slots(x_SVT, x_ncol,
+						slotp, sloti, slotx);
 		if (ret < 0) {
 			UNPROTECT(3);
 			error("SparseArray internal error in "
@@ -664,26 +598,68 @@ SEXP C_from_SVT_SparseMatrix_to_CsparseMatrix(SEXP x_dim,
 		}
 	}
 
-	ans = PROTECT(NEW_LIST(3));
-	SET_VECTOR_ELT(ans, 0, ans_p);
-	SET_VECTOR_ELT(ans, 1, ans_i);
-	SET_VECTOR_ELT(ans, 2, ans_x);
-	UNPROTECT(4);
+	SEXP ans = PROTECT(NEW_LIST(3));
+	SET_VECTOR_ELT(ans, 0, slotp);
+	SET_VECTOR_ELT(ans, 1, sloti);
+	SET_VECTOR_ELT(ans, 2, slotx);
+	UNPROTECT(drop_slotx ? 3 : 4);
 	return ans;
 }
 
 
 /****************************************************************************
- * Going from [d|l]gCMatrix to SVT_SparseMatrix
+ * Going from [d|l|n]gCMatrix to SVT_SparseMatrix
  */
 
-/* Returns R_NilValue or a "leaf vector" of length <= nzcount.
+static char get_gCMatrix_subtype(SEXP x)
+{
+	SEXP class_attrib = GET_CLASS(x);
+	const char *class = CHAR(STRING_ELT(class_attrib, 0));
+	if (strcmp(class, "dgCMatrix") == 0)
+		return 'd';
+	if (strcmp(class, "lgCMatrix") == 0)
+		return 'l';
+	if (strcmp(class, "ngCMatrix") == 0)
+		return 'n';
+	error("'x' must be a [d|l|n]gCMatrix object");
+}
+
+static SEXP build_leaf_from_ngCsparseMatrix_col(SEXP x_sloti,
+		int ix_offset, int col_nzcount, SEXPTYPE ans_Rtype)
+{
+	SEXP ans_nzoffs = PROTECT(NEW_INTEGER(col_nzcount));
+	memcpy(INTEGER(ans_nzoffs), INTEGER(x_sloti) + ix_offset,
+	       sizeof(int) * col_nzcount);
+	SEXP ans_nzvals = LACUNAR_MODE_IS_ON ?
+		R_NilValue : PROTECT(_new_Rvector1(ans_Rtype, col_nzcount));
+	SEXP ans = zip_leaf(ans_nzvals, ans_nzoffs);
+	UNPROTECT(LACUNAR_MODE_IS_ON ? 1 : 2);
+	return ans;
+}
+
+/* Returns R_NilValue or a leaf with an nzcount <= 'col_nzcount'.
 
    Note that, quite surprisingly, and unfortunately, [d|l]gCMatrix objects
    can sometimes have zeros in their "x" slot.
    For example, with Matrix 1.3-4:
 
        dgcm <- as(matrix(c(11:13, 0, 0, 21:25), ncol=2), "dgCMatrix")
+
+    In an lgCMatrix object:
+
+       lgcm <- dgcm >= 13
+       lgcm
+       # 5 x 2 sparse Matrix of class "lgCMatrix"
+       # [1,] : |
+       # [2,] : |
+       # [3,] | |
+       # [4,] . |
+       # [5,] . |
+       lgcm@x
+       # [1] FALSE FALSE  TRUE  TRUE  TRUE  TRUE  TRUE  TRUE
+
+    In a dgCMatrix object:
+
        dgcm[cbind(3:4, 2)] <- 0
        dgcm
        # 5 x 2 sparse Matrix of class "dgCMatrix"
@@ -707,30 +683,32 @@ SEXP C_from_SVT_SparseMatrix_to_CsparseMatrix(SEXP x_dim,
        dgcm@x
        # [1] 13 21 22 25
 
-   We want to make sure that these zeros don't end up in the "leaf vector"
-   returned by build_leaf_vector_from_CsparseMatrix_col(). Unfortunately,
-   this introduces an additional cost to coercion from [d|l]gCMatrix to
-   SVT_SparseMatrix. This cost is a slowdown that is (approx.) between 1.3x
-   and 1.5x. */
-static SEXP build_leaf_vector_from_CsparseMatrix_col(SEXP x_i, SEXP x_x,
-		int offset, int nzcount,
-		SEXPTYPE ans_Rtype, int *warn, int *offs_buf)
+   We want to make sure that these zeros don't end up in the leaf returned
+   by build_leaf_from_CsparseMatrix_col(). Unfortunately, this introduces
+   an additional cost to coercion from [d|l]gCMatrix to SVT_SparseMatrix.
+   This cost is a slowdown that is (approx.) between 1.3x and 1.5x. */
+static SEXP build_leaf_from_CsparseMatrix_col(SEXP x_sloti, SEXP x_slotx,
+		int ix_offset, int col_nzcount,
+		SEXPTYPE ans_Rtype, int *warn, int *nzoffs_buf)
 {
-	SEXP ans, ans_offs;
-	int ans_len;
-
-	/* Will skip zeros from 'x_x' if any. */
-	ans = _make_leaf_vector_from_Rsubvec(x_x, (R_xlen_t) offset, nzcount,
-					     offs_buf, 1);
+	/* 'x_slotx' can contain zeros. See above. */
+	SEXP ans = _make_leaf_from_Rsubvec(x_slotx,
+				(R_xlen_t) ix_offset, col_nzcount,
+				nzoffs_buf, 1);
 	if (ans == R_NilValue)
 		return ans;
+
 	PROTECT(ans);
-	ans_offs = VECTOR_ELT(ans, 0);
-	ans_len = LENGTH(ans_offs);  /* can only be <= 'nzcount' */
-	_copy_selected_ints(INTEGER(x_i) + offset, INTEGER(ans_offs), ans_len,
-			    INTEGER(ans_offs));
-	if (ans_Rtype != TYPEOF(x_x))
-		ans = _coerce_leaf_vector(ans, ans_Rtype, warn, offs_buf);
+
+	/* Replace offsets in 'ans_nzoffs' with offsets from 'x_sloti'. */
+	SEXP ans_nzoffs = get_leaf_nzoffs(ans);
+	int ans_nzcount = LENGTH(ans_nzoffs);  /* always <= 'col_nzcount' */
+	_copy_selected_int_elts(INTEGER(x_sloti) + ix_offset,
+				INTEGER(ans_nzoffs), ans_nzcount,
+				INTEGER(ans_nzoffs));
+	if (ans_Rtype != TYPEOF(x_slotx))
+		ans = _coerce_leaf(ans, ans_Rtype, warn, nzoffs_buf);
+
 	UNPROTECT(1);
 	return ans;
 }
@@ -738,43 +716,44 @@ static SEXP build_leaf_vector_from_CsparseMatrix_col(SEXP x_i, SEXP x_x,
 /* --- .Call ENTRY POINT --- */
 SEXP C_build_SVT_from_CsparseMatrix(SEXP x, SEXP ans_type)
 {
-	SEXPTYPE ans_Rtype;
-	SEXP x_Dim, x_p, x_i, x_x, ans, ans_elt;
-	int x_nrow, x_ncol, j, offset, nzcount, warn, *offs_buf, is_empty;
-
-	ans_Rtype = _get_Rtype_from_Rstring(ans_type);
+	char x_type = get_gCMatrix_subtype(x);
+	SEXPTYPE ans_Rtype = _get_Rtype_from_Rstring(ans_type);
 	if (ans_Rtype == 0)
 		error("invalid requested type");
 
-	x_Dim = GET_SLOT(x, install("Dim"));
-	x_nrow = INTEGER(x_Dim)[0];
-	x_ncol = INTEGER(x_Dim)[1];
-	x_p = GET_SLOT(x, install("p"));
+	SEXP x_Dim = GET_SLOT(x, install("Dim"));
+	int x_nrow = INTEGER(x_Dim)[0];
+	int x_ncol = INTEGER(x_Dim)[1];
+	SEXP x_slotp = GET_SLOT(x, install("p"));
 
-	if (INTEGER(x_p)[x_ncol] == 0)
+	if (INTEGER(x_slotp)[x_ncol] == 0)
 		return R_NilValue;
 
-	x_i = GET_SLOT(x, install("i"));
-	x_x = GET_SLOT(x, install("x"));
+	SEXP x_sloti = GET_SLOT(x, install("i"));
+	SEXP x_slotx = x_type == 'n' ? R_NilValue : GET_SLOT(x, install("x"));
 
-	offs_buf = (int *) R_alloc(x_nrow, sizeof(int));
-	ans = PROTECT(NEW_LIST(x_ncol));
-	warn = 0;
-	is_empty = 1;
-	for (j = 0; j < x_ncol; j++) {
-		offset = INTEGER(x_p)[j];
-		nzcount = INTEGER(x_p)[j + 1] - offset;
-		if (nzcount != 0) {
-			ans_elt = build_leaf_vector_from_CsparseMatrix_col(
-						x_i, x_x,
-						offset, nzcount,
-						ans_Rtype, &warn, offs_buf);
-			if (ans_elt != R_NilValue) {
-				PROTECT(ans_elt);
-				SET_VECTOR_ELT(ans, j, ans_elt);
-				UNPROTECT(1);
-				is_empty = 0;
-			}
+	int *nzoffs_buf = (int *) R_alloc(x_nrow, sizeof(int));
+	SEXP ans = PROTECT(NEW_LIST(x_ncol));
+	int warn = 0;
+	int is_empty = 1;
+	for (int j = 0; j < x_ncol; j++) {
+		int ix_offset = INTEGER(x_slotp)[j];
+		int col_nzcount = INTEGER(x_slotp)[j + 1] - ix_offset;
+		if (col_nzcount == 0)
+			continue;
+		SEXP ans_elt = x_type == 'n' ?
+			build_leaf_from_ngCsparseMatrix_col(x_sloti,
+					ix_offset, col_nzcount, ans_Rtype
+			)
+			:
+			build_leaf_from_CsparseMatrix_col(x_sloti, x_slotx,
+					ix_offset, col_nzcount,
+					ans_Rtype, &warn, nzoffs_buf);
+		if (ans_elt != R_NilValue) {
+			PROTECT(ans_elt);
+			SET_VECTOR_ELT(ans, j, ans_elt);
+			UNPROTECT(1);
+			is_empty = 0;
 		}
 	}
 	if (warn)
@@ -790,9 +769,7 @@ SEXP C_build_SVT_from_CsparseMatrix(SEXP x, SEXP ans_type)
 
 static SEXP alloc_nzvals(SEXP type, R_xlen_t n)
 {
-	SEXPTYPE Rtype;
-
-	Rtype = _get_Rtype_from_Rstring(type);
+	SEXPTYPE Rtype = _get_Rtype_from_Rstring(type);
 	if (Rtype == 0)
 		error("SparseArray internal error in alloc_nzvals():\n"
 		      "    SVT_SparseArray object has invalid type");
@@ -803,21 +780,18 @@ static SEXP alloc_nzvals(SEXP type, R_xlen_t n)
 SEXP C_from_SVT_SparseArray_to_COO_SparseArray(SEXP x_dim,
 		SEXP x_type, SEXP x_SVT)
 {
-	R_xlen_t nzcount;
-	SEXP nzcoo, nzvals, ans;
-
-	nzcount = _REC_nzcount_SVT(x_SVT, LENGTH(x_dim));
+	R_xlen_t nzcount = _REC_nzcount_SVT(x_SVT, LENGTH(x_dim));
 	if (nzcount > INT_MAX)
 		error("SVT_SparseArray object contains too many nonzero "
 		      "values to be turned into a COO_SparseArray object");
-
-	nzvals = PROTECT(alloc_nzvals(x_type, nzcount));
-	nzcoo = PROTECT(extract_nzcoo_and_nzvals_from_SVT(x_SVT,
-					(int) nzcount, LENGTH(x_dim), nzvals));
-
-	ans = PROTECT(NEW_LIST(2));
-	SET_VECTOR_ELT(ans, 0, nzcoo);
-	SET_VECTOR_ELT(ans, 1, nzvals);
+	SEXP ans_nzvals = PROTECT(alloc_nzvals(x_type, nzcount));
+	SEXP ans_nzcoo = PROTECT(
+		extract_nzcoo_and_nzvals_from_SVT(x_SVT,
+				(int) nzcount, LENGTH(x_dim), ans_nzvals)
+	);
+	SEXP ans = PROTECT(NEW_LIST(2));
+	SET_VECTOR_ELT(ans, 0, ans_nzcoo);
+	SET_VECTOR_ELT(ans, 1, ans_nzvals);
 	UNPROTECT(3);
 	return ans;
 }
